/*********************************************************************************
 * port_filestream.c ($Revision: 189 $)
 *
 * Copyright (c) 2006-2008 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "port_filestream.c"

// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Open and close input-/output file for audio streaming (CAPI/VOIP)
// When clockrate is given, the file is treated as stream (named pipe / fifo)
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
pj_status_t port_filestream_player_recorder(connection_t *mCON, char* in_file, char* out_file, int clockrate, int speex, long buffer_size, int wait_start, int wait_end, int *play_id)
{
    connection_t *CUR_CON = (connection_t*)mCON;    
    pj_status_t status = PJ_SUCCESS;    
    int i;

    // play_id NULL?? Error!
    if(play_id == NULL)
        return 2;

    // Speex encoder? then use fix clockrate!
    if(speex)
        clockrate = 8000;

    // Input file-stream
    if(strcmp(in_file, ""))
    {
        // Init resampler
        if(strcmp(in_file, DEVNULL)) {        
            status = port_filestream_init(CUR_CON, in_file, "", clockrate, speex, buffer_size, wait_start, wait_end, play_id);
        }
        // Stop resampler
        else 
        {   
            // close last play_queue (-stop play)
            // close play_queue with play_id (-stop play [play_id])
            if (*play_id >= -1) {
                // close specified player
                if (*play_id >= INCONF_PORTS)
                  *play_id = -1;
                port_filestream_close(CUR_CON, *play_id, 1);           
            } else if (*play_id == -2) {
                // close ALL players
                for (i=0; i<INCONF_PORTS; i++)
                    port_filestream_close(CUR_CON, i, 1);
            } else if (*play_id < -2) {
                // close last player
                port_filestream_close(CUR_CON, -1, 1);
            }
         
            status = 1;
        }
    }

    // Output file-stream
    if(strcmp(out_file, ""))
    {
        // Close file rec port, if opened
        if(CUR_CON->ports->rec_resample != NULL && CUR_CON->ports->rec_stream != NULL && CUR_CON->slots->rec != -1) {
            port_filestream_close(CUR_CON, -1, 0);                        
            
        }
        //else {
        if(CUR_CON->ports->rec_resample == NULL && CUR_CON->ports->rec_stream == NULL && CUR_CON->slots->rec == -1) {
            // Init resampler
            if(strcmp(out_file, DEVNULL))         
                status = port_filestream_init(CUR_CON, "", out_file, clockrate, speex, buffer_size, wait_start, wait_end, play_id);
        }    
    }
        
    return status;
}

// Initialize filestream and resample port
static pj_status_t port_filestream_init(connection_t *mCON, char *in_file, char *out_file, int clockrate, int speex, long buffer_size, int wait_start, int wait_end, int *play_id)
{    
    connection_t *CUR_CON = (connection_t*)mCON;
    portdata_filestream *pdata;
    pj_status_t status = PJ_SUCCESS;
    int queue = -1;    
    int queue_cnt = 0;
    int wait_cnt=0;

    // initialize play_id with -1 (=error)
    // on success, the correct play_id (=queue_id) is set
    *play_id = -1; 

    // Wait 250 msec. Maybe connection is not active, yet.
    // Otherwise, abort!
    while(CUR_CON->slots->media == -1 && wait_cnt < 25) {        
        pj_thread_sleep(10);
        wait_cnt++;
    }    
    if(CUR_CON->slots->media == -1) {        
        return PJMEDIA_ERROR;
    }

    // --------------------------------------------------------------------------------------------
    // Player
    // --------------------------------------------------------------------------------------------
    if(strcmp(in_file, "") && strcmp(in_file, DEVNULL))
    {
        // Get free playing queue
        queue = play_by_free(CUR_CON->id, &queue_cnt);        
        if(queue == -1)
        {
            app_err(THIS_FILE, "No free playing queue left. Max: %d!", INCONF_PORTS);            
            return 1;
        }
        
        // When currently nothings is played, reset memory pool 
        if(queue_cnt == 0)
        {               
            // Destroy old connection memory pool (if exists)
            if(CUR_CON->pools->play != NULL) {
              pj_pool_release(CUR_CON->pools->play);
              CUR_CON->pools->play = NULL;
            }

            // Create own memory pool for play-streaming
            if(CUR_CON->pools->play == NULL)
                CUR_CON->pools->play = pj_pool_create(&oGLB.cp.factory, "streaming_pool_play (%p)", POOL_SIZE_CON, POOL_INCR_CON, NULL);
        }
        
        // Play raw data from stream or wave file
        status = port_filestream_create(CUR_CON->pools->play, &CUR_CON->ports->play_stream[queue], in_file, out_file, clockrate, speex, buffer_size, wait_start, wait_end, CUR_CON->id, queue);
        if (status != PJ_SUCCESS) 
        {
            app_err(THIS_FILE, "(%d) Unable to open %s", status, in_file);
            
            port_filestream_close(CUR_CON, queue, 1);
            return status;
        }

        // Get private port data
        pdata = CUR_CON->ports->play_stream[queue]->port_data.pdata;
        
        // pjmedia_wav_player_port specific
        if(!clockrate)
        {            
           if(CUR_CON->ports->play_stream[queue] != NULL)
           {                
                // Check channel count (wave-file only)
                if (pdata->file_wave_play_port->info.channel_count != CONF_CHANNELS) {
                    app_err(THIS_FILE, "File has different number of channels. Only 1 channel allowed (mono)!");                    
                                        
                    port_filestream_close(CUR_CON, queue, 1);
                    return 1;
                }

                // Create a resample port for CAPI and VoIP!
                status = pjmedia_resample_port_create(CUR_CON->pools->play, pdata->file_wave_play_port, oSET.audio.conf_clockrate, oSET.audio.resampler_quality, &CUR_CON->ports->play_resample[queue]);
                if (status != PJ_SUCCESS) {                                
                    app_err(THIS_FILE, "(%d) Unable to create resample port", status);                    
                    
                    port_filestream_close(CUR_CON, queue, 1);
                    return status;
                }       
           }
        }
        else 
        {
            // Create a resample port for CAPI and VoIP!
            if(CUR_CON->ports->play_stream[queue] != NULL && CUR_CON->ports->play_resample[queue] == NULL)
            {
                status = pjmedia_resample_port_create(CUR_CON->pools->play, CUR_CON->ports->play_stream[queue], oSET.audio.conf_clockrate, oSET.audio.resampler_quality, &CUR_CON->ports->play_resample[queue]);
                if (status != PJ_SUCCESS) {                                
                    app_err(THIS_FILE, "(%d) Unable to create resample port", status);                    
                    
                    port_filestream_close(CUR_CON, queue, 1);
                    return status;
                }                            
            }
        }
     
        // Add play_port to conference and connect ports        
        if(CUR_CON->ports->play_resample[queue] != NULL)
        {
            char c_name[32];
            pj_str_t name;

            pj_ansi_snprintf(c_name, sizeof(c_name), "player [%d,%d]", CUR_CON->id, queue);
            name = pj_str(c_name);

            // Hook resampler callback (getframe)
            pdata->resampled_cb_get_frame = CUR_CON->ports->play_resample[queue]->get_frame;        
            CUR_CON->ports->play_resample[queue]->get_frame = &port_filestream_resampled_get_frame;                
         
            status = pjmedia_conf_add_port(oGLB.conf.bridge, CUR_CON->pools->play, CUR_CON->ports->play_resample[queue], &name, (unsigned int*)&CUR_CON->slots->play[queue]);
            if (status != PJ_SUCCESS) {                    
                app_err(THIS_FILE, "(%d) Unable to add wave-player to conference bridge!", status);
                
                port_filestream_close(CUR_CON, queue, 1);
                return status;        
            }

            // Connect play_slot with media_slot
            status = pjmedia_conf_connect_port(oGLB.conf.bridge, CUR_CON->slots->play[queue], CUR_CON->slots->media, 0);
        } 

        // set "currently playing" parameters
        if(status == PJ_SUCCESS)
        {
            if(sh_actions != NULL) {
                sh_actions->play_rec.play_thread[CUR_CON->id][queue] = queue;
                sh_actions->retval.play_id = queue;
            }

            // set play_id = queue_id
            *play_id = queue;
        } 
        else 
        {                        
            port_filestream_close(CUR_CON, queue, 1);
        }
    }

    // --------------------------------------------------------------------------------------------
    // Recorder
    // --------------------------------------------------------------------------------------------
    if(strcmp(out_file, "") && strcmp(out_file, DEVNULL))
    {   
        // Destroy old connection memory pool (if exists)
        if(CUR_CON->pools->rec != NULL) {
            pj_pool_release(CUR_CON->pools->rec);
            CUR_CON->pools->rec = NULL;
        }
       
        // Create own memory pool for play-streaming
        if(CUR_CON->pools->rec == NULL)
            CUR_CON->pools->rec = pj_pool_create(&oGLB.cp.factory, "streaming_pool_rec (%p)", POOL_SIZE_CON, POOL_INCR_CON, NULL);

        // Record to stream or wave file
        status = port_filestream_create(CUR_CON->pools->rec, &CUR_CON->ports->rec_stream, in_file, out_file, clockrate, speex, buffer_size, wait_start, wait_end, CUR_CON->id, 0);        
        if (status != PJ_SUCCESS) 
        {
            app_err(THIS_FILE, "(%d) Unable to open %s", status, out_file);
            
            port_filestream_close(CUR_CON, 0, 0);       
            return status;
        }

        // Get private port data
        pdata = CUR_CON->ports->rec_stream->port_data.pdata;
        
        // Check channel count
        if(!clockrate)
        {
           if(pdata->file_wave_rec_port != NULL)
           {
                // Wave-File must have same number of channels. 
                if (pdata->file_wave_rec_port->info.channel_count != CONF_CHANNELS) {
                    app_err(THIS_FILE, "File has different number of channels. Only 1 channel allowed (mono)!");
                    
                    port_filestream_close(CUR_CON, 0, 0);   
                    return 1;
                }

                // Create a resample port
                status = pjmedia_resample_port_create(CUR_CON->pools->rec, pdata->file_wave_rec_port, oSET.audio.conf_clockrate, oSET.audio.resampler_quality, &CUR_CON->ports->rec_resample);
                if (status != PJ_SUCCESS) 
                {            
                    app_err(THIS_FILE, "(%d) Unable to create resample port", status);
                    
                    port_filestream_close(CUR_CON, 0, 0);   
                    return status;
                }
           }
        }
        else
        {
            // Create a resample port         
            if(CUR_CON->ports->rec_stream != NULL)
            {
                status = pjmedia_resample_port_create(CUR_CON->pools->rec, CUR_CON->ports->rec_stream, oSET.audio.conf_clockrate, oSET.audio.resampler_quality, &CUR_CON->ports->rec_resample);
                if (status != PJ_SUCCESS) 
                {            
                    app_err(THIS_FILE, "(%d) Unable to create resample port", status);
                    
                    port_filestream_close(CUR_CON, 0, 0);   
                    return status;
                }
            }   
        }

        // Add file_rec_port to conference and connect ports
        if(CUR_CON->ports->rec_resample != NULL)
        {            
            char c_name[32];
            pj_str_t name;

            pj_ansi_snprintf(c_name, sizeof(c_name), "recorder [%d]", CUR_CON->id);
            name = pj_str(c_name);

            // Hook resampler callback (putframe)
            pdata->resampled_cb_put_frame = CUR_CON->ports->rec_resample->put_frame;        
            CUR_CON->ports->rec_resample->put_frame = &port_filestream_resampled_put_frame;                

            // Add wave_recorder port to conference bridge
            status = pjmedia_conf_add_port(oGLB.conf.bridge, CUR_CON->pools->rec, CUR_CON->ports->rec_resample, &name, (unsigned int*)&CUR_CON->slots->rec);
            if (status != PJ_SUCCESS) {                        
                app_err(THIS_FILE, "(%d) Unable to add wave-recorder to conference bridge!", status);
                
                port_filestream_close(CUR_CON, 0, 0);   
                return status;
            }
            
            status = pjmedia_conf_connect_port(oGLB.conf.bridge, CUR_CON->slots->media, CUR_CON->slots->rec, 0);                    
            if(status == PJ_SUCCESS)
            {                                  
                *play_id = -1;
            } 
            else 
            {
                port_filestream_close(CUR_CON, 0, 0);   
                return status;
            }
        }     
        else
        {            
            port_filestream_close(CUR_CON, 0, 0);   
            return 1;
        }
    }
    
    return status;
}

// Create filestream port
static pj_status_t port_filestream_create(pj_pool_t *pool, pjmedia_port **p_port, char *in_file, char *out_file, int clockrate, int speex, long buffer_size, int wait_start, int wait_end, int con_id, int play_id)
{
    pjmedia_port *port;
    portdata_filestream *pdata;    
    unsigned count;    
    int is_wave_port = 0;
    pj_status_t status = PJ_SUCCESS;

    PJ_ASSERT_RETURN(pool, PJ_EINVAL);

    port = pj_pool_zalloc(pool, sizeof(pjmedia_port));
    PJ_ASSERT_RETURN(port != NULL, PJ_ENOMEM);

    // wavefile or stream?
    if(clockrate == 0) {
        is_wave_port = 1;
        clockrate = 8000;
    }

    // speex file?
    if(speex)
    {
        is_wave_port = 0;
        clockrate = 8000;
    }

    // default buffersize, when not set
    if(buffer_size <= 0 || buffer_size >= 262144) {
        buffer_size = ((clockrate * 20 / 1000) * (CONF_BITS / 8)) * N_FILESTREAM_BUFFERS;
    }

    // default timeout in frames (wait_start / wait_end). 1 sec.    
    if((wait_start <= 0 || wait_start > 1000) && wait_start != -1)
        wait_start = 50;   
    if((wait_end <= 0 || wait_end > 1000) && wait_end != -1)
        wait_end = 50;     

    // fill port info
    port->info.bits_per_sample = CONF_BITS;
    port->info.channel_count = CONF_CHANNELS;
    port->info.encoding_name = pj_str("pcm");
    port->info.has_info = 1;
    port->info.name = pj_str("filestream port");
    port->info.need_info = 0;
    port->info.pt = 0xFF;
    port->info.clock_rate = clockrate;
    port->info.samples_per_frame = (clockrate * 20 / 1000);
    port->info.bytes_per_frame = port->info.samples_per_frame * (CONF_BITS / 8);
    port->info.type = PJMEDIA_TYPE_AUDIO;

    // Create portdata 
    port->port_data.pdata = pdata = pj_pool_zalloc(pool, sizeof(portdata_filestream));
    
    // Initialize portdata
    count = port->info.samples_per_frame / CONF_CHANNELS;
    pdata->buf_size = port->info.bytes_per_frame * (int)(buffer_size / (long)port->info.samples_per_frame); //N_FILESTREAM_BUFFERS;
    pdata->buf = pj_pool_alloc(pool, pdata->buf_size);
    PJ_ASSERT_RETURN(pdata->buf != NULL, PJ_ENOMEM);    
    pdata->buf_write = 0;
    pdata->buf_read = 0;
    pdata->stream_started = 0;

    // save additional port_info data
    pdata->con_id = con_id;
    pdata->play_id = play_id;    
    pdata->wait_start = (wait_start * -1);
    pdata->wait_end = wait_end;
    if(wait_start == -1) pdata->wait_start = 1;  // endless?
    if(wait_end == -1) pdata->wait_end = -1;     // endless?
    pdata->eof_count = 0;
    pdata->mixdata.tx_volume = vol_lin2log(vol_scale(oSET.audio.mixer_tx_volume)); 
    pdata->mixdata.rx_volume = vol_lin2log(vol_scale(oSET.audio.mixer_rx_volume)); 
    pdata->mixdata.mother_port = port;  

    // Init speex encoder/decoder
    #if DTMFBOX_SPEEX_ENCODER != 0    
        if(speex)
            speex_converter_init(&pdata->speex);
    #endif

    // on_destroy() callback
    port->on_destroy = &port_filestream_on_destroy;

    #if DTMFBOX_WIN32
        pdata->fp_in = NULL;
        pdata->fp_out = NULL;
    #else
        pdata->fp_in = -1;
        pdata->fp_out = -1;
    #endif

    // Stream or Speex?
    if(!is_wave_port)
    {
        // file reader
        if(strcmp(in_file, "") && strcmp(in_file, DEVNULL))
        {
            #if DTMFBOX_WIN32
              pdata->fp_in = fopen(in_file, "rb");
              if(pdata->fp_in == NULL)                    
            #else        
              if((pdata->fp_in = open(in_file, O_NONBLOCK | O_RDONLY, 0)) == -1)
            #endif           
              {
                *p_port = port;
                return 1;
              }

            // get_frame() callback
            port->get_frame = &port_filestream_get_frame;
            if(speex)
                app_dbg2(THIS_FILE, "Speex player created (%s, %dhz, %dbit, mono)", in_file, clockrate, CONF_BITS);
            else
                app_dbg2(THIS_FILE, "Stream player created (%s, %dhz, %dbit, mono)", in_file, clockrate, CONF_BITS);
        }

        // file writer
        if(strcmp(out_file, "") && strcmp(out_file, DEVNULL))
        {
            #if DTMFBOX_WIN32
              pdata->fp_out = fopen(out_file, "wb");
              if(pdata->fp_out == NULL)              
            #else            
              if((pdata->fp_out = open(out_file, O_WRONLY, 0)) == -1)              
            #endif        
              {   
                *p_port = port;
                return 1;
              }

            // put_frame() callback
            port->put_frame = &port_filestream_put_frame;            
            if(speex)
                app_dbg2(THIS_FILE, "Speex recorder created (%s, %dhz, %dbit, mono)", out_file, clockrate, CONF_BITS);
            else
                app_dbg2(THIS_FILE, "Stream recorder created (%s, %dhz, %dbit, mono)", out_file, clockrate, CONF_BITS);
        }
    }
    
    else
    {        
        pj_str_t remote_ip;
        char target_addr[MAX_URI_LENGTH];
        char *port;
        int port_is_numeric = 0;
        pjmedia_dir dir = PJMEDIA_DIR_NONE;
        int play = 0;
        int rec = 0;        
        
        // Playing?
        if(strcmp(in_file, "") && strcmp(in_file, DEVNULL)) {
            strncpy(target_addr, in_file, sizeof(target_addr));
            play = 1;
            dir += PJMEDIA_DIR_DECODING; 
        }

        // Recording?
        if(strcmp(out_file, "") && strcmp(out_file, DEVNULL)) {
            strncpy(target_addr, out_file, sizeof(target_addr));
            rec = 1;
            dir += PJMEDIA_DIR_ENCODING; 
        }

        // RTP/RTCP ?
        remote_ip = pj_str(strtok(target_addr, ":")); 
        port = strtok(NULL, ":");   
        if(port != NULL)
            port_is_numeric = isnumeric(port);

        if(port != NULL && port_is_numeric) 
        {
            pj_sockaddr_in remote_addr;
            pj_uint16_t remote_port = 4000;
            pj_uint16_t rtp_port = 5000;            
            if(isnumeric(port)) remote_port = (pj_uint16_t)atoi(port);
            if(sh_actions->play_rec.rtp_port != 0) rtp_port = (pj_uint16_t)sh_actions->play_rec.rtp_port;
            
            status = pj_sockaddr_in_init(&remote_addr, &remote_ip, remote_port);

            if(status == PJ_SUCCESS)
            {
                if(play)
                    status =  port_rtpstream_init(pool, dir, rtp_port, &remote_addr, con_id, play_id, &pdata->file_wave_play_port);
                
                if(rec)
                    status =  port_rtpstream_init(pool, dir, rtp_port, &remote_addr, con_id, -1, &pdata->file_wave_rec_port);
                
            }
        } 
        // WAVE ?
        else 
        {
            // Player
            if(play) 
                status = pjmedia_wav_player_port_create(pool, in_file, 0, PJMEDIA_FILE_NO_LOOP, 0, &pdata->file_wave_play_port);            
            
            // Recorder
            if(rec)
                status = pjmedia_wav_writer_port_create(pool, out_file, oSET.audio.conf_clockrate, CONF_CHANNELS, oSET.audio.conf_samplerate, CONF_BITS, 0, 0, &pdata->file_wave_rec_port);         
        }
    }

    *p_port = port;

    return status;
}

// Close filestream port
// play_queue = -1 (close last queue), >= 0 (close specified queue)
// close_playport = 1 playing, 0 = recording
static pj_status_t port_filestream_close(connection_t *mCON, int play_queue, int close_playport)
{    
    connection_t *CUR_CON = (connection_t*)mCON;           
    pj_status_t status = PJ_SUCCESS;
    int queue = 0;
    int found_queue = 0;
        
    // close player
    if(close_playport)
    {          
        // get last playing queue        
        if(play_queue == -1)        
        {
            for(queue=INCONF_PORTS-1; queue>=0; queue--)
            {
                if(CUR_CON->ports->play_stream[queue] != NULL && CUR_CON->ports->play_resample[queue] != NULL && CUR_CON->slots->play[queue] != -1)
                {
                    found_queue = 1;
                    break;
                }
            }

            if(found_queue == 0) {                          
                return 1;         
            }
        }
        // get specified playing queue
        else 
        {
            if(CUR_CON->ports->play_stream[play_queue] != NULL || CUR_CON->ports->play_resample[play_queue] != NULL || CUR_CON->slots->play[play_queue] != -1)
            {
                queue = play_queue;
            }
            else
            {                
                if(sh_actions != NULL)
                    sh_actions->play_rec.play_thread[CUR_CON->id][play_queue] = -1;  
                
                return 1;
            }
        }

        // remove play_slot from conference
        if(CUR_CON->slots->play[queue] != -1) 
        {   
            // Remove play_slot from conference             
            mixer_conf_port_disconnect_all(CUR_CON->slots->play[queue]);
            pjmedia_conf_remove_port(oGLB.conf.bridge, CUR_CON->slots->play[queue]);                                               
            CUR_CON->slots->play[queue] = -1;            
        }   

        // destroy resample port
        if(CUR_CON->ports->play_resample[queue] != NULL) {
            status = pjmedia_port_destroy(CUR_CON->ports->play_resample[queue]);            
            CUR_CON->ports->play_resample[queue] = NULL;              
        }

        // destroy stream port
        if(CUR_CON->ports->play_stream[queue] != NULL) 
        {
            status = pjmedia_port_destroy(CUR_CON->ports->play_stream[queue]);                                 
            CUR_CON->ports->play_stream[queue] = NULL;                                
        }
        
        // tell that playing stopped (stop blocking)!                           
        if(sh_actions != NULL)
            sh_actions->play_rec.play_thread[CUR_CON->id][queue] = -1;                

        /*
        // the last playing queue? Reset memory pool!
        {
            int queue_cnt;
            int queue_tmp;
            queue_tmp = play_by_free(CUR_CON->id, &queue_cnt);
            if(queue_cnt == 0 && CUR_CON->pools->play) {
                pj_pool_release(CUR_CON->pools->play);
                CUR_CON->pools->play = NULL;
            }
        }
        */
    }
   

    // close recorder
    else
    {   
          // remove from conference 
        if(CUR_CON->slots->rec != -1)
        {            
            mixer_conf_port_disconnect_all(CUR_CON->slots->rec);                                                                        
            status = pjmedia_conf_remove_port(oGLB.conf.bridge, CUR_CON->slots->rec);                               
            CUR_CON->slots->rec = -1;
        } 

        // destroy record resampler
        if(CUR_CON->ports->rec_resample != NULL) 
        {
            status = pjmedia_port_destroy(CUR_CON->ports->rec_resample);            
            CUR_CON->ports->rec_resample = NULL;              
        }

        // destroy rec stream port
        if(CUR_CON->ports->rec_stream != NULL)
        {
            status = pjmedia_port_destroy(CUR_CON->ports->rec_stream);            
            CUR_CON->ports->rec_stream = NULL; 
        }
    }
    
    return status;
}

// destroy callback for filestream port
static pj_status_t port_filestream_on_destroy(pjmedia_port *port)
{      
    pj_status_t status = PJ_SUCCESS;
    portdata_filestream *pdata = (portdata_filestream*)port->port_data.pdata;
    
    // close input file (-playstream / -playstreamthread only)
    #if DTMFBOX_WIN32
     if(pdata->fp_in != NULL) {
        fclose(pdata->fp_in);                    
     }
     if(pdata->fp_out != NULL) {
        fclose(pdata->fp_out);                    
     }
    #else
     if(pdata->fp_in != -1) {
        close(pdata->fp_in);                    
        pdata->fp_in = -1;
     }
     if(pdata->fp_out != -1) {
        close(pdata->fp_out);                    
        pdata->fp_out = -1;
     }
    #endif   
   
    // reset "hook"
    if(pdata->resampled_cb_get_frame)                
       pdata->resampled_cb_get_frame = NULL;

    if(pdata->resampled_cb_put_frame) {
        pdata->resampled_cb_put_frame = NULL;
    }

    // Destroy speex encoder/decoder
    #if DTMFBOX_SPEEX_ENCODER != 0    
        speex_converter_close(&pdata->speex);
    #endif

    return status;
}

// We cannot destroy a filestream_port in port_filestream_eof. 
// This is because we are in a callback and selfdestroy isn't that good ;)
// We call this function in signal_handler_thread() periodically and check, 
// if the stream/file got an EOF. When EOF is detected, we close the port.
pj_status_t port_filestream_close_all_eof()
{
    int i, j;

    // walkthrough connections
    for(i=0; i<oSET.max_connections; i++) 
    {
        if(oCONS[i].info->type != FREE)
        {
            // walkthrough file_play_port and check if EOF is reached.
            for(j=0; j<INCONF_PORTS; j++) {
                if(sh_actions->play_rec.play_thread[i][j] == -1 && (oCONS[i].ports->play_stream[j] != NULL && oCONS[i].ports->play_resample[j] != NULL && oCONS[i].slots->play[j] != -1))
                    port_filestream_close(&oCONS[i], j, 1);
            }
        }
    }

    return PJ_SUCCESS;
}

// Player EOF callback 
static pj_status_t port_filestream_eof(pjmedia_port *port, void* usr_data)
{
    portdata_filestream *pdata = port->port_data.pdata;    
    pj_status_t status = PJ_SUCCESS;
    int play_queue = 0;
    int found_queue = 0;

    PJ_UNUSED_ARG(usr_data);

    // search for the queue
    for(play_queue=0; play_queue<INCONF_PORTS; play_queue++)
    {
        if(oCONS[pdata->con_id].ports->play_stream[play_queue] != NULL)
        {
            portdata_filestream *pdata2 = oCONS[pdata->con_id].ports->play_stream[play_queue]->port_data.pdata;
            
            // stream?
            if(port == oCONS[pdata->con_id].ports->play_stream[play_queue]) {
                found_queue = 1;
                break;
            }
            
            // wave file?
            if(port == pdata2->file_wave_play_port) {
                found_queue = 1;
                break;
            }
        }
    }    
    if(!found_queue)
        play_queue = -1;

    
    // maz: never destroy the filestream_port directly!
    // close filestream    
    // port_filestream_close(&oCONS[pdata->con_id], play_queue, 1);


    // maz: tell second instance that playing stopped (stop blocking)!
    //      The file_port is finally closed by port_filestream_close_all_eof() in signal_handler_thread().     
    if(sh_actions != NULL && found_queue) {
       if(sh_actions->play_rec.play_thread[pdata->con_id][play_queue] != -1)
         sh_actions->play_rec.play_thread[pdata->con_id][play_queue] = -1;                                
    }

    return status;
}

// Put raw data to stream(-file)
static pj_status_t port_filestream_put_frame(pjmedia_port *port, const pjmedia_frame *frame)
{    
    portdata_filestream *pdata = port->port_data.pdata;
    pj_int16_t *samples = frame->buf;         
    pj_status_t status = PJ_SUCCESS;
      
    //#if DTMFBOX_WIN32

    if(pdata->fp_out)  
    {
        #if DTMFBOX_SPEEX_ENCODER != 0            
            if(pdata->speex.enc)
            {
                // SPEEX:
                pjmedia_frame frame_out;
                frame_out.buf = frame->buf;        
                speex_encode_frame(&pdata->speex, frame, &frame_out);

                #if DTMFBOX_WIN32
                    fwrite(&frame_out.size, 1, 1, pdata->fp_out);
                    fwrite(frame_out.buf, frame_out.size, 1, pdata->fp_out);
                #else
                    fd_set wfd;
                    int len;            
                    FD_ZERO(&wfd);
                    FD_SET(pdata->fp_out, &wfd);                
                    len = write(pdata->fp_out, &frame_out.size, 1);
                    len = write(pdata->fp_out, frame_out.buf, frame_out.size);        
                #endif
            }        
            else
        #endif
            {     
                // STREAM/RAW:
                #if DTMFBOX_WIN32
                    fwrite(samples, frame->size, 1, pdata->fp_out);
                #else
                    fd_set wfd;
                    int len;            
                    FD_ZERO(&wfd);
                    FD_SET(pdata->fp_out, &wfd);
                    len = write(pdata->fp_out, samples, frame->size);
                #endif
            }
    }
    else
    {
        status = PJ_EEOF;
    }

    return status;
}



// Get raw data from stream(-file)
static pj_status_t port_filestream_get_frame(pjmedia_port *port, pjmedia_frame *frame)
{
    portdata_filestream *pdata = port->port_data.pdata;
    pj_int16_t *samples = frame->buf;     
    pj_status_t status = PJ_SUCCESS;
    
    frame->type = PJMEDIA_FRAME_TYPE_AUDIO;
    frame->timestamp.u64 = 0;

    #if DTMFBOX_WIN32
    // TODO: apply input ring buffer to windows code ? 
    //  or is fread's buffering enough?
    if(pdata->fp_in != NULL)
    {
        if(!feof(pdata->fp_in)) 
        {                       
            #if DTMFBOX_SPEEX_ENCODER != 0    

                if(pdata->speex.dec)
                {
                    // SPEEX:            
                    pjmedia_frame frame_out;                            
                    int bytes = 38;
                    
                    fread(&bytes, 1, 1, pdata->fp_in);                                        
                    while(bytes != 38 && !feof(pdata->fp_in))
                        fread(&bytes, 1, 1, pdata->fp_in);                                                            

                    if(!feof(pdata->fp_in))
                        fread(samples, bytes, 1, pdata->fp_in);        

                    frame->size = bytes;
                    frame_out.size = bytes;                           
                    frame_out.buf = frame->buf;
                    
                    speex_decode_frame(&pdata->speex, frame, &frame_out);                        
                }                
                else

            #endif
                {
                    // STREAM/RAW:
                    fread(samples, frame->size, 1, pdata->fp_in);
                }
        }
        else 
        {
            pj_bzero(frame->buf, frame->size);
            status = 1;  
        }
    }
    else
    {
        pj_bzero(frame->buf, frame->size);
        status = 1; 
    }
    
    #else
    // the linux code has a filestream input buffer
    {        
        #if DTMFBOX_SPEEX_ENCODER != 0    

            if(pdata->speex.dec)
            {
                // SPEEX: 
                pjmedia_frame frame_out;                            
                int bytes = 38;
                int len = 0;
                
                len = read(pdata->fp_in, &bytes, 1);                  
                while(bytes != 38 && len > 0)
                    len = read(pdata->fp_in, &bytes, 1);

                frame->size = bytes;
                frame_out.size = bytes;                           
                frame_out.buf = frame->buf;

                if(len > 0)
                {
                    len = read(pdata->fp_in, samples, bytes);        
                    status = speex_decode_frame(&pdata->speex, frame, &frame_out);                                        
                } 
                else 
                {
                    status = PJ_EEOF;
                }
            }            
            else
        #endif
            {
                // RAW/STREAM:
                status = port_filestream_get_buffer(port, frame);
                port_filestream_fill_buffer(port);
            }
    }
    #endif

    // EOF?
    if(status) 
    {
        pj_bzero(frame->buf, frame->size);
        frame->type = PJMEDIA_FRAME_TYPE_NONE;
        frame->size = 0;
    }
    
    return status;
}

// Hooked resample-port callback (calls mixer_port_put_frame)
static pj_status_t port_filestream_resampled_put_frame(pjmedia_port *port, const pjmedia_frame *frame)
{
    portdata_filestream *pdata = NULL;
    pj_int16_t *samples = frame->buf;
    pjmedia_frame frame_new;
    pj_status_t status = PJ_SUCCESS;
    int i;

    frame_new.buf = samples;
    frame_new.size = frame->size;
    frame_new.type = frame->type;
    frame_new.timestamp = frame->timestamp;
    frame_new.type = PJMEDIA_FRAME_TYPE_AUDIO;

    // search for private port_data 
    for(i=0; i<oSET.max_connections; i++) {
        if(oCONS[i].info->type != FREE)
        {
            if(oCONS[i].ports->rec_resample == port && oCONS[i].ports->rec_stream != NULL)
            {
                pdata = oCONS[i].ports->rec_stream->port_data.pdata;
                break;
            }   
        }
    }
        
    if(pdata != NULL) 
    {       
        // Mixer callback
        if(port)
            mixer_port_put_frame(port, &frame_new, pdata->con_id, 0);

        pdata = oCONS[i].ports->rec_stream->port_data.pdata;

        // Real resample callback
        if(pdata->resampled_cb_put_frame && oCONS[i].ports->rec_resample)            
            return pdata->resampled_cb_put_frame(port, &frame_new);
        else {            
            status = PJ_EEOF;
        }

    } else {
        status = PJ_EEOF;
    }

    return status;
}

// Hooked resample-port callback (calls mixer_port_get_frame)
static pj_status_t port_filestream_resampled_get_frame(pjmedia_port *port, pjmedia_frame *frame)
{
    portdata_filestream *pdata = NULL;
    pj_int16_t *samples = frame->buf;
    pjmedia_frame frame_new;
    pj_status_t status = PJ_SUCCESS;
    int i;
    int k;
    
    frame_new.buf = samples;    
    frame_new.size = frame->size;
    frame_new.type = frame->type;
    frame_new.timestamp = frame->timestamp;
    
    // search for private port_data 
    for(i=0; i<oSET.max_connections; i++) {
        if(oCONS[i].info->type != FREE)
        {
            for(k=0; k<INCONF_PORTS; k++) {
                if(oCONS[i].ports->play_resample[k] == port && oCONS[i].ports->play_stream[k] != NULL)
                {
                    pdata = oCONS[i].ports->play_stream[k]->port_data.pdata;
                    break;
                }
            }
            if(pdata != NULL)
                break;
        }
    }
        
    if(pdata != NULL) 
    {        
        // Call real resample callback (get the frames)
        if(pdata->resampled_cb_get_frame && port)
            status = pdata->resampled_cb_get_frame(port, &frame_new);
        else
            status = PJ_EEOF;

        // EOF?
        if(status != PJ_SUCCESS) 
        {            
            //if(pdata != NULL) 
            {
                //pdata = oCONS[i].ports->play_stream[k]->port_data.pdata;

                // When stream/file error occurs, wait and send 0 byte for silence. 
                // After that, close the stream/file.            
                if(pdata->eof_count == N_FILESTREAM_BUFFERS) 
                {            
                    // EOF received! 
                    port_filestream_eof(oCONS[pdata->con_id].ports->play_stream[pdata->play_id], NULL);                
                }
                else
                {
                    // count eof and send 0-bytes (silence)!
                    pj_bzero(frame_new.buf, frame_new.size);                    
                    pdata->eof_count++;     
                    status = PJ_SUCCESS;
                }
            }
        }

        // Mixer callback
        if(status == PJ_SUCCESS) {
            mixer_port_get_frame(oCONS[pdata->con_id].ports->play_stream[pdata->play_id], &frame_new, pdata->con_id, pdata->play_id, status);    
            frame->type = PJMEDIA_FRAME_TYPE_AUDIO;
        }
        else {
            pj_bzero(frame_new.buf, frame_new.size);                    
            frame_new.type = PJMEDIA_FRAME_TYPE_NONE;
            frame_new.size = 0;            
        }
    }
          
    return status;
}

/* 

helper functions for input ringbuffer (Linux/Unix only) 

thx flo :)

*/

#if(DTMFBOX_WIN32 == 0)

// convert continuous buffer index to real ring buffer index
static __inline unsigned long warp_index(unsigned long i, unsigned size)
{
  return (i%size);
}

// read a frame from file <fp_in> in ringbuffer buf, 
//  return updated ring buffer write index wr
static unsigned long read_in_ringbuffer(int fp_in, pj_int16_t *buf, unsigned size, unsigned long wr, unsigned long rd)
{    
    unsigned long free = size - (wr-rd);
    if ( free > 0 ) {
    unsigned i1 = warp_index(wr,size);
    unsigned l1 = size - i1;
    unsigned i2=0;
    unsigned l2=0;
    int l1r=0, l2r=0;

    if (l1>free) { l1 = free; }
    else {
      i2 = 0;
      l2 = free-l1;
    }

    if (l1>0) 
      l1r = read(fp_in, buf+i1/(sizeof(pj_int16_t)), l1);
    if ((l2>0)&&(l1r==l1))
      l2r = read(fp_in, buf+i2/(sizeof(pj_int16_t)), l2);

    // error: ignore, but set 0 bytes read (not -1, -2 .. error codes!!)
    if (l1r < 0) l1r = 0;
    if (l2r < 0) l2r = 0;

    return wr+(l1r+l2r);
    }

    return wr;
}

// copy a frame from ringbuffer to output buffer *outbuf of size outsize, 
//  return updated ring buffer read index rd
static unsigned long copy_out_ringbuffer(pj_int16_t *buf, unsigned size, unsigned long wr, unsigned long rd, pj_int16_t *outbuf, unsigned outsize)
{    
    unsigned long filled = wr-rd;
    if ( filled >= outsize ) {

    unsigned i1 = warp_index(rd,size);
    unsigned l1 = size - i1;
    unsigned i2=0;
    unsigned l2=0;

    if (l1>=outsize) { l1 = outsize; }
    else {
      i2 = 0;
      l2 = outsize-l1;
    }
    pj_memcpy(outbuf, buf+i1/(sizeof(pj_int16_t)), l1);
    if (l2)
      pj_memcpy(outbuf+l1/(sizeof(pj_int16_t)), buf+i2/(sizeof(pj_int16_t)), l2);

    return rd+l1+l2;
    }
    return rd;
}

// copy and remove a pjmedia frame from the ringbuffer
static pj_status_t port_filestream_get_buffer(pjmedia_port *port, pjmedia_frame *frame)
{    
    portdata_filestream *pdata = port->port_data.pdata;
    pj_status_t status = PJ_SUCCESS;

    unsigned long filled = pdata->buf_write - pdata->buf_read;

    // case 1: there is data in buffer to return
    if (filled >= frame->size) 
    {
        // FIXME: output this only at a higher loglevel!
        //if (pdata->stream_started != 1) 
        //  app_dbg(THIS_FILE,"first stream data received!\n");

        pdata->stream_started = 1;
        pdata->stream_ended = -1;
        pdata->buf_read = copy_out_ringbuffer(pdata->buf, pdata->buf_size, pdata->buf_write, pdata->buf_read, (pj_int16_t *)frame->buf, frame->size);                
    } 

    // case 2: not enough data in buffer
    else if ((filled > 0)&&(pdata->stream_started)) {
        
        // return what we can
        pdata->buf_read = copy_out_ringbuffer(pdata->buf, pdata->buf_size, pdata->buf_write, pdata->buf_read, (pj_int16_t *)frame->buf, filled);
        
        // pad the rest with zeros
        pj_bzero((pj_int16_t *) frame->buf + filled/(sizeof(pj_int16_t)), frame->size - filled);
        
        // reset wait_end, when data is available         
        pdata->stream_ended = -1;
    }

    // case 3: buffer completely empty
    else {
        // FIXME: output this only at a higher loglevel!
        //app_dbg(THIS_FILE,"<WARNING> no data in input buffer!\n");

        pj_bzero(frame->buf, frame->size);
        if (pdata->stream_started == 1) // EOF, buffer underrun
        {
            // wait forever (wait_end = -1)?
            if(pdata->wait_end == -1) {
                pdata->stream_ended = -1;                
                status = PJ_SUCCESS;
            }
            else {
                // or wait timeout (wait_end > 0)?
                pdata->stream_ended++;
                if (pdata->stream_ended > pdata->wait_end) { 
                    // assume EEOF if no data received within buffer_size 
                    status = PJ_EEOF;
                }
            }
        } else {
            // wait forever (wait_start = -1)?
            if(pdata->wait_start == 1) {                                
                status = PJ_SUCCESS;
            }
            else
            {
                // or wait timeout (wait_start > 0)?
                pdata->stream_started--;
                if (pdata->stream_started < (pdata->wait_start)) {
                    // return EEOF if no data received within 50 frames after port was created
                    status = PJ_EEOF;
                }
            }
        }
    }

    return status;
}

// fill the input ringbuffer by reading as much data as possible from input filestream
static pj_status_t port_filestream_fill_buffer(pjmedia_port *port)
{
    portdata_filestream *pdata = port->port_data.pdata;
    struct timeval timeout;
    fd_set rfd;
    int ready;
    pj_status_t status = PJ_SUCCESS;

    timeout.tv_sec = 0;    // no timeout
    timeout.tv_usec = 0;
 
    if(pdata->fp_in >= 0)
    {
      unsigned long free;

      // prevent overflow of pointers.... after couple of hours ;-)
      if (pdata->buf_write > (1<<30)) {
        long tmp = pdata->buf_read - pdata->buf_size;
        pdata->buf_write -= tmp;
        pdata->buf_read -= tmp;
      }

      free = pdata->buf_size - (pdata->buf_write - pdata->buf_read);

      //app_dbg(THIS_FILE,"fill_buffer: bs: %i -> buf_wr: %i , buf_rd %i ,free: %i\n",pdata->buf_size,pdata->buf_write, pdata->buf_read,free);

      if (free > pdata->buf_size/2) {
        FD_ZERO(&rfd);
        FD_SET(pdata->fp_in, &rfd);

        ready = select(pdata->fp_in+1, &rfd, NULL, NULL, &timeout);
        if(ready >= 0) 
        {
           pdata->buf_write = read_in_ringbuffer(pdata->fp_in, pdata->buf, 
                                  pdata->buf_size, pdata->buf_write, pdata->buf_read);
        }
        // else: if no data read, we still have the buffer ;-)
        //    EEOF is now only returned if the buffer is empty
      }
    }

    return status;        
}

#endif
