/*********************************************************************************
 * port_rtpstream.c ($Revision: 189 $)
 *
 * Copyright (c) 2006-2008 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "port_rtpstream.c"

pj_status_t port_rtpstream_init(pj_pool_t *pool, pjmedia_dir dir, pj_uint16_t local_port, const pj_sockaddr_in *rem_addr, int con_id, int play_id, pjmedia_port **p_port)
{   
    pjmedia_port *port;    
    pjmedia_transport *transport;
    const pjmedia_codec_info *codec_info;
    portdata_rtpstream *pdata;
    pj_status_t status = PJ_SUCCESS;
        
    PJ_ASSERT_RETURN(oGLB.pool, PJ_EINVAL);
    PJ_UNUSED_ARG(play_id);

    // Initialize this port
    port = pj_pool_zalloc(pool, sizeof(pjmedia_port));
    PJ_ASSERT_RETURN(port != NULL, PJ_ENOMEM);

    // Create rtpstream pdata 
    port->port_data.pdata = pdata = pj_pool_zalloc(pool, sizeof(portdata_rtpstream));
    
    // Fill port info    
    port->info.bits_per_sample = CONF_BITS;
    port->info.channel_count = CONF_CHANNELS;
    port->info.encoding_name = pj_str("pcm");
    port->info.has_info = 1;
    port->info.name = pj_str("rtpstream port");
    port->info.need_info = 0;
    port->info.pt = 0xFF;
    port->info.clock_rate = oSET.audio.conf_clockrate;
    port->info.samples_per_frame = port->info.clock_rate * 20 / 1000;
    port->info.bytes_per_frame = port->info.samples_per_frame * (CONF_BITS / 8);
    port->info.type = PJMEDIA_TYPE_AUDIO;

    // Initialize callbacks
    port->put_frame = &port_rtpstream_put_frame;
    port->get_frame = &port_rtpstream_get_frame;
    port->on_destroy = &port_rtpstream_on_destroy;

    // Fill private port data
    //pdata->con_id = CUR_CON->id;   
    pdata->mixdata.tx_volume = vol_lin2log(vol_scale(oSET.audio.mixer_tx_volume));
    pdata->mixdata.rx_volume = vol_lin2log(vol_scale(oSET.audio.mixer_rx_volume));  
    pdata->mixdata.mother_port = port;
    pdata->con_id = con_id;

    // Reset stream info.
    pj_bzero(&pdata->info, sizeof(pdata->info));

    // Initialize stream info formats 
    pdata->info.type = PJMEDIA_TYPE_AUDIO;
    pdata->info.dir = dir;
	pjmedia_codec_mgr_get_codec_info(pjmedia_endpt_get_codec_mgr(oGLB.med_endpt), 0, &codec_info);   // Use the first initialized codec as default
    pj_memcpy(&pdata->info.fmt, codec_info, sizeof(pjmedia_codec_info));
    pdata->info.tx_pt = codec_info->pt;    
    pdata->info.ssrc = pj_rand();
        
    // Copy remote address 
    pj_memcpy(&pdata->info.rem_addr, rem_addr, sizeof(pj_sockaddr_in));

    // Create media transport
    status = pjmedia_transport_udp_create(oGLB.med_endpt, NULL, local_port, 0, &transport);
    if (status != PJ_SUCCESS) {	            
        return status;
    }

    // Create media stream
    status = pjmedia_stream_create(oGLB.med_endpt, pool, &pdata->info, transport, NULL, &pdata->rtp_stream_port);
    if (status != PJ_SUCCESS) {	            
        app_err(THIS_FILE, "(%d) Cannot create RTP stream", status);
        return status;
    }

    // Get the port interface of the stream 
    status = pjmedia_stream_get_port(pdata->rtp_stream_port, &pdata->rtp_media_port);
    if (status != PJ_SUCCESS) {	            
        app_err(THIS_FILE, "(%d) Cannot get media port from RTP stream", status);
        return status;
    }
    
    // Start streaming 
    pjmedia_stream_start(pdata->rtp_stream_port);

    *p_port = port;

    return status;
}

static pj_status_t port_rtpstream_on_destroy(pjmedia_port *port)
{
    portdata_rtpstream *pdata = port->port_data.pdata;

    // Destroy stream port
    if(pdata->rtp_stream_port)
    {
        pjmedia_transport *tp;

        tp = pjmedia_stream_get_transport(pdata->rtp_stream_port);
	    pjmedia_stream_destroy(pdata->rtp_stream_port);
	    pjmedia_transport_udp_close(tp);
    }

    return PJ_SUCCESS;
}

static pj_status_t port_rtpstream_get_frame(pjmedia_port *port, pjmedia_frame *frame)
{
    portdata_rtpstream *pdata = port->port_data.pdata;
    pj_int16_t *samples = frame->buf;
    pjmedia_frame frame_new;
    pj_status_t status = PJ_SUCCESS;
            
    frame_new.buf = samples;
    frame_new.size = frame->size;
    frame_new.type = frame->type;
    frame_new.timestamp = frame->timestamp;

    // Get frames from stream port
    status = pjmedia_port_get_frame(pdata->rtp_media_port, &frame_new);          
    
    // Mixer callback
    /*
    if(port)
    {
        pdata = port->port_data.pdata;
        mixer_port_get_frame(port, &frame_new, pdata->con_id, -1, status);
    }
    */
    return status;
}

static pj_status_t port_rtpstream_put_frame(pjmedia_port *port, const pjmedia_frame *frame)
{
    portdata_rtpstream *pdata = port->port_data.pdata;
    pj_int16_t *samples = frame->buf;
    pjmedia_frame frame_new;

    frame_new.buf = samples;
    frame_new.size = frame->size;
    frame_new.type = frame->type;
    frame_new.timestamp = frame->timestamp;

    // Mixer callback        
    // mixer_port_put_frame(port, &frame_new, pdata->con_id, -1);

    // Send frames to stream port
    pdata = port->port_data.pdata;  
    return pjmedia_port_put_frame(pdata->rtp_media_port, &frame_new);
}
