/*********************************************************************************
 * port_tone.c ($Revision: 189 $)
 *
 * Copyright (c) 2006-2008 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "port_tone.c"

pj_status_t port_tone_init(connection_t *mCON, short freq1, short freq2, short on_msec, short off_msec, short volume)
{
    connection_t *CUR_CON = (connection_t*)mCON;    
    pj_status_t status = PJ_SUCCESS;
    
    PJ_ASSERT_RETURN(CUR_CON != NULL, 1);    
    PJ_ASSERT_RETURN(CUR_CON->ports->tone == NULL, 2);    
    
    status = port_tone_create(CUR_CON->pools->con_pool, CUR_CON, &CUR_CON->ports->tone, freq1, freq2, on_msec, off_msec, volume);
    if(status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to create tone port!", status);
        port_tone_close(CUR_CON);
    }

    return status;
}

pj_status_t port_tone_close(connection_t *mCON)
{
    connection_t *CUR_CON = (connection_t*)mCON;        
    portdata_tone *pdata;
    pj_status_t status = PJ_SUCCESS;
    
    PJ_ASSERT_RETURN(CUR_CON != NULL, 1);    
    PJ_ASSERT_RETURN(CUR_CON->ports->tone != NULL, 2);    

    // disconnect this tone_slot from every connection
    mixer_conf_port_disconnect_all(CUR_CON->slots->tone);

    // remote tone slot from conference
    if(CUR_CON->slots->tone != -1) {
        status = pjmedia_conf_remove_port(oGLB.conf.bridge, CUR_CON->slots->tone);        
        CUR_CON->slots->tone = -1;
    }

    // destroy tone port
    if(CUR_CON->ports->tone != NULL)
    {
        // destroy tone generator
        pdata = CUR_CON->ports->tone->port_data.pdata;    
        if(pdata->tonegen_port != NULL)
            status = pjmedia_port_destroy(pdata->tonegen_port);

        status = pjmedia_port_destroy(CUR_CON->ports->tone);
        if(status != PJ_SUCCESS)
            app_err(THIS_FILE, "(%d) Unable to destroy tone port!", status);
        else
            CUR_CON->ports->tone = NULL;
    }

    return status;
}

pj_status_t port_tone_create(pj_pool_t *pool, connection_t *mCON, pjmedia_port **p_port, short freq1, short freq2, short on_msec, short off_msec, short volume)
{   
    connection_t *CUR_CON = (connection_t*)mCON;    
    pjmedia_port *port;
    portdata_tone *pdata;    
    pj_status_t status = PJ_SUCCESS;
    
    PJ_ASSERT_RETURN(pool, PJ_EINVAL);

    port = pj_pool_zalloc(pool, sizeof(pjmedia_port));
    PJ_ASSERT_RETURN(port != NULL, PJ_ENOMEM);

    // Fill port info
    port->info.bits_per_sample = CONF_BITS;
    port->info.channel_count = CONF_CHANNELS;
    port->info.encoding_name = pj_str("pcm");
    port->info.has_info = 1;
    port->info.name = pj_str("tone port");
    port->info.need_info = 0;
    port->info.pt = 0xFF;
    port->info.clock_rate = oSET.audio.conf_clockrate;
    port->info.samples_per_frame = port->info.clock_rate * 20 / 1000; 
    port->info.bytes_per_frame = port->info.samples_per_frame * (CONF_BITS / 8);
    port->info.type = PJMEDIA_TYPE_AUDIO;

    // Initialize callbacks    
    port->get_frame = &port_tone_get_frame;

    // Create capi_port pdata 
    port->port_data.pdata = pdata = pj_pool_zalloc(pool, sizeof(portdata_tone));
    
    // Fill port data
    pdata->freq1 = freq1;
    pdata->freq2 = freq2;
    pdata->off_msec = off_msec;
    pdata->on_msec = on_msec;
    pdata->volume = volume;
    pdata->con_id = CUR_CON->id;    

    // Create tone generator
    status = pjmedia_tonegen_create(CUR_CON->pools->con_pool, oSET.audio.conf_clockrate, CONF_CHANNELS, oSET.audio.conf_samplerate, CONF_BITS, 0, &pdata->tonegen_port);
    if(status != PJ_SUCCESS) {
       app_err(THIS_FILE, "(%d) Unable to create tone generator!", status);        
    }

    // Add tone port to conference
    if(status == PJ_SUCCESS && oGLB.conf.bridge != NULL && CUR_CON->slots->tone == -1)
    {
        char c_name[32];
        pj_str_t name;

        pj_ansi_snprintf(c_name, sizeof(c_name), "tone port [%d]", CUR_CON->id);
        name = pj_str(c_name);
        status = pjmedia_conf_add_port(oGLB.conf.bridge, pool, port, &name, (unsigned int*)&CUR_CON->slots->tone);
        pj_assert(status == PJ_SUCCESS);
    }

    *p_port = port;

    return status;
}


void port_tone_ring_thread(void** mCONS)
{      
    connection_t *oCONsrc = &oCONS[(int)mCONS[0]];    
    connection_t *oCONdst = &oCONS[(int)mCONS[1]];     
    pj_status_t status;

    // no connection? return!
    if(oCONsrc == NULL || oCONdst == NULL)
        return;

    // no media_slot, no earlyb3_ring ??? return!!!
    if(oCONsrc->slots->media == -1 || !oSET.audio.earlyb3_ring)
        return;

    // Create tone-port
    if(oCONdst->ports->tone == NULL)
    {
        status = port_tone_create(oCONdst->pools->con_pool, oCONdst, &oCONdst->ports->tone, TONE_PROGRESS[0], TONE_PROGRESS[1], TONE_PROGRESS[2], TONE_PROGRESS[3], TONE_PROGRESS[4]);
        if(status != PJ_SUCCESS)
            return;
    }

    // Connect tone-port to the connection, that is requesting the call
    status = pjmedia_conf_connect_port(oGLB.conf.bridge, oCONdst->slots->tone, oCONsrc->slots->media, 0);
    
    if(status == PJ_SUCCESS)
    {
        // Wait until connection status changes
        while((oCONdst->info->active == 0 && oCONdst->info->status == 1) &&
             ((oCONsrc->info->active == 1 && oCONsrc->info->status == 2) || (oCONsrc->info->type == SCRD)))
             {   
                  pj_thread_sleep(25);
             }

        // Stop tone-playing 
        if(oCONdst->ports->tone)
        {
            portdata_tone *pdata = oCONdst->ports->tone->port_data.pdata;

            if(pdata->tonegen_port != NULL)
            {
                pjmedia_tonegen_stop(pdata->tonegen_port);

                while(pjmedia_tonegen_is_busy(pdata->tonegen_port))
                    pj_thread_sleep(25);

                pjmedia_tonegen_stop(pdata->tonegen_port);
            }
        }
    }

    // Close tone_port    
    if(oCONdst->ports->tone)
        port_tone_close(oCONdst);                  
}

// Tone-Creator
static pj_status_t port_tone_generate_tones(pjmedia_port *p_port)
{   
    pjmedia_tone_desc tones[PJMEDIA_TONEGEN_MAX_DIGITS];
    portdata_tone *pdata = p_port->port_data.pdata;    
    pj_status_t status = PJ_SUCCESS;
    int i;

    if(pdata->tonegen_port != NULL)
    {
        if(!pjmedia_tonegen_is_busy(pdata->tonegen_port))
        {           
            pjmedia_tonegen_stop(pdata->tonegen_port);

            for(i=0; i<PJMEDIA_TONEGEN_MAX_DIGITS; i++)
            {
                tones[i].freq1 = pdata->freq1;
                tones[i].freq2 = pdata->freq1;
                tones[i].on_msec = pdata->on_msec;
                tones[i].off_msec = pdata->off_msec;
                tones[i].volume = pdata->volume;
            }

            status = pjmedia_tonegen_play(pdata->tonegen_port, PJMEDIA_TONEGEN_MAX_DIGITS, tones, 0);                                         
        }
    }

    return status;
}

static pj_status_t port_tone_get_frame(pjmedia_port *port, pjmedia_frame *frame)
{
    portdata_tone *pdata = port->port_data.pdata;
    pj_status_t status = PJ_SUCCESS;    
        
    // generate tones 
    status = port_tone_generate_tones(port);

    // receive generated tones from downstream port 
    if(status == PJ_SUCCESS) {
        status = pjmedia_port_get_frame(pdata->tonegen_port, frame);                                    
    }
    else {
        frame->type = PJMEDIA_FRAME_TYPE_NONE;
        pj_memset(frame->buf, 0, frame->size);        
    }
        
    return status;
}

