/*********************************************************************************
 * port_voip.c ($Revision: 189 $)
 *
 * Copyright (c) 2006-2008 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "port_voip.c"

pj_status_t port_voip_init(connection_t *mCON)
{
    connection_t *CUR_CON = (connection_t*)mCON;    
    pj_status_t status = PJ_SUCCESS;
    
    PJ_ASSERT_RETURN(CUR_CON != NULL, 1);    
        
    // Create custom_port and hook media_port callbacks...
    status = port_voip_create(CUR_CON->pools->con_pool, CUR_CON, &CUR_CON->ports->custom);
    if(status != PJ_SUCCESS)    
        app_err(THIS_FILE, "(%d) Unable to create VoIP port!", status);
    
    return status;
}

pj_status_t port_voip_close(connection_t *mCON)
{
    connection_t *CUR_CON = (connection_t*)mCON;    
    portdata_voip *pdata;
    pj_status_t status = PJ_SUCCESS;

    PJ_ASSERT_RETURN(CUR_CON != NULL, 1);    
    PJ_ASSERT_RETURN(CUR_CON->ports->custom != NULL, 2);    
    
    // get portdata
    pdata = (portdata_voip*)CUR_CON->ports->custom->port_data.pdata;
       
    // remove media slots from conference bridge
    if(CUR_CON->slots->media != -1)
    {        
        // Disconnect audioless loopback
        pjmedia_conf_disconnect_port(oGLB.conf.bridge, CUR_CON->slots->media, oGLB.audio.flow_null_slot);
        pjmedia_conf_disconnect_port(oGLB.conf.bridge, oGLB.audio.flow_null_slot, CUR_CON->slots->media);

        // disconnect from each other media_slot
        mixer_conf_port_disconnect_all(CUR_CON->slots->media);

        // remove media_slot from conference
        pjmedia_conf_remove_port(oGLB.conf.bridge, CUR_CON->slots->media);        
        CUR_CON->slots->media = (unsigned)-1;
    } 
    else 
    {
        // no valid media_slot? then just remove from in_conf[] array
        int i;
        for(i=0;i<oSET.max_connections; i++)
            mixer_conf_array_remove(&oCONS[i], CUR_CON->id);        
    }

    // free media port
    if(CUR_CON->ports->media != NULL)
    {
        pjmedia_port_destroy(CUR_CON->ports->media);
        CUR_CON->ports->media = NULL;
    }

    // destroy custom voip_port
    if(CUR_CON->ports->custom != NULL)
    {
        status = pjmedia_port_destroy(CUR_CON->ports->custom);
        if(status != PJ_SUCCESS)    
            app_err(THIS_FILE, "(%d) Unable to destroy VoIP port!", status);
        else
            CUR_CON->ports->custom = NULL;
    }
    
    return status;
}

static pj_status_t port_voip_create(pj_pool_t *pool, connection_t *mCON, pjmedia_port **p_port)
{   
    connection_t *CUR_CON = (connection_t*)mCON;    
    pjmedia_port *port;
    portdata_voip *pdata = NULL;    
    pj_status_t status = PJ_SUCCESS;
    unsigned count;    
    
    PJ_ASSERT_RETURN(pool, PJ_EINVAL);
    
    port = *p_port;

    if(port == NULL)
    {
        // port_voip_init() was called the first time. 
        // Prepare cust_port...

        port = pj_pool_zalloc(pool, sizeof(pjmedia_port));
        PJ_ASSERT_RETURN(port != NULL, PJ_ENOMEM);
        
        // Fill port info
        port->info.bits_per_sample = CONF_BITS;
        port->info.channel_count = CONF_CHANNELS;
        port->info.encoding_name = pj_str("pcm");
        port->info.has_info = 1;
        port->info.name = pj_str("voip port");
        port->info.need_info = 0;
        port->info.pt = 0xFF;
        port->info.clock_rate = oSET.audio.conf_clockrate;
        port->info.samples_per_frame = port->info.clock_rate * 20 / 1000; 
        port->info.bytes_per_frame = port->info.samples_per_frame * (CONF_BITS / 8);
        port->info.type = PJMEDIA_TYPE_AUDIO;

        // Initialize callbacks
        port->put_frame = &port_voip_put_frame;
        port->get_frame = &port_voip_get_frame;
        
        // Create voip_port pdata 
        port->port_data.pdata = pdata = pj_pool_zalloc(pool, sizeof(portdata_voip));
        
        // Fill voip_port pdata
        count = port->info.samples_per_frame / CONF_CHANNELS;
        pdata->con_id = CUR_CON->id;                    
        pdata->mixdata.tx_volume = vol_lin2log(vol_scale(oSET.audio.mixer_tx_volume));
        pdata->mixdata.rx_volume = vol_lin2log(vol_scale(oSET.audio.mixer_rx_volume));
        pdata->mixdata.mother_port = port;
    }

    // port_voip_init() was called a second time (or first).         
    if(CUR_CON->ports->media != NULL)
    {   
        // Get private port data
        if(CUR_CON->ports->custom != NULL)
            pdata = CUR_CON->ports->custom->port_data.pdata;
        
        if(pdata != NULL)
        {
            char c_name[32];
            pj_str_t name;

            pj_ansi_snprintf(c_name, sizeof(c_name), "voip port [%d]", CUR_CON->id);
            name = pj_str(c_name);

            // Add media_port to conference
            status = pjmedia_conf_add_port(oGLB.conf.bridge, CUR_CON->pools->con_pool, CUR_CON->ports->media, &name, (unsigned int*)&CUR_CON->slots->media);
            pj_assert(status == PJ_SUCCESS);

            // Hook the media_port callbacks..            
            pdata->mediaport_cb_get_frame = CUR_CON->ports->media->get_frame;
            pdata->mediaport_cb_put_frame = CUR_CON->ports->media->put_frame;
            CUR_CON->ports->media->get_frame = &port_voip_get_frame;
            CUR_CON->ports->media->put_frame = &port_voip_put_frame;

            // Connect audioless loopback
            pjmedia_conf_connect_port(oGLB.conf.bridge, CUR_CON->slots->media, oGLB.audio.flow_null_slot, 0);
            pjmedia_conf_connect_port(oGLB.conf.bridge, oGLB.audio.flow_null_slot, CUR_CON->slots->media, 0);
        }
        else
        {
            status = PJMEDIA_ERROR;
        }
    }

    *p_port = port;

    return status;
}

static pj_status_t port_voip_get_frame(pjmedia_port *port, pjmedia_frame *frame)
{    
    pj_int16_t *samples = frame->buf;    
    portdata_voip *pdata = NULL;
    pjmedia_frame frame_new;    
    pj_status_t status = PJ_SUCCESS;
    int i;
    
    frame_new.buf = samples;
    frame_new.size = frame->size;
    frame_new.type = frame->type;
    frame_new.timestamp = frame->timestamp;

    // Search for cust_port (must be media_port) and get pdata
    for(i=0; i<oSET.max_connections; i++) {
        if(port == oCONS[i].ports->media && oCONS[i].ports->media != NULL) {
            pdata = oCONS[i].ports->custom->port_data.pdata;        
            break;
        }
    }  
    if(pdata != NULL) 
    {
        // Call real media_port getframe()
        status = pdata->mediaport_cb_get_frame(port, &frame_new);        
        if(status == PJ_SUCCESS)
            frame->type = PJMEDIA_FRAME_TYPE_AUDIO;

        // Mixer callback
        mixer_port_get_frame(oCONS[i].ports->custom, &frame_new, pdata->con_id, -1, status);
    }

    return status;
}

static pj_status_t port_voip_put_frame(pjmedia_port *port, const pjmedia_frame *frame)
{    
    pj_int16_t *samples = frame->buf;    
    portdata_voip *pdata = NULL;
    pjmedia_frame frame_new;    
    pj_status_t status = PJ_SUCCESS;            
    int i; 

    frame_new.buf = samples;
    frame_new.size = frame->size;
    frame_new.type = frame->type;
    frame_new.timestamp = frame->timestamp;

    // Search for cust_port (must be media_port) and get pdata
    for(i=0; i<oSET.max_connections; i++) {
        if(port == oCONS[i].ports->media && oCONS[i].ports->media != NULL) {
            pdata = oCONS[i].ports->custom->port_data.pdata;        
            break;
        }
    }  

    if(pdata != NULL)
    {     
        // Mixer callback
        mixer_port_put_frame(oCONS[i].ports->custom, &frame_new, pdata->con_id, -1);

        // Call real media_port putframe()
        return pdata->mediaport_cb_put_frame(port, &frame_new);
    }

    return status;
}
