/*********************************************************************************
 * speex.c ($Revision: 189 $)
 *
 * Copyright (c) 2006-2008 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "speex.c"

#if DTMFBOX_SPEEX_ENCODER != 0

pj_status_t speex_converter_init(speex_converter_data *spx)
{
    pj_status_t status = PJ_SUCCESS;

    #if defined(PJMEDIA_HAS_SPEEX_CODEC) && PJMEDIA_HAS_SPEEX_CODEC!=0

    // Init encoder
    spx->clockrate = CONF_CLOCKRATE;
    spx->enc_mode = &speex_nb_mode;
    spx->dec_mode = &speex_nb_mode;
    spx->penh = 0;

    spx->enc = speex_encoder_init(spx->enc_mode);
    if(!spx->enc)
    {
        app_err(THIS_FILE, "Unable to initialize speex encoder!");
        return PJMEDIA_CODEC_EFAILED;
    }
    speex_bits_init(&spx->enc_bits);

    // set quality
    //speex_encoder_ctl(enc, SPEEX_SET_QUALITY, 1);

    // set clockrate
    speex_encoder_ctl(spx->enc, SPEEX_SET_SAMPLING_RATE, &spx->clockrate);

    // VAD 
    spx->vad = (oSET.voip.use_vad != 0);
    speex_encoder_ctl(spx->enc, SPEEX_SET_VAD, &spx->vad);
    speex_encoder_ctl(spx->enc, SPEEX_SET_DTX, &spx->vad);

    // Init decoder
    spx->dec = speex_decoder_init(spx->dec_mode);
    if(!spx->dec)
    {
        app_err(THIS_FILE, "Unable to initialize speex decoder!");
        return PJMEDIA_CODEC_EFAILED;
    }
    speex_bits_init(&spx->dec_bits);

    // set clockrate
    speex_encoder_ctl(spx->dec, SPEEX_SET_SAMPLING_RATE, &spx->clockrate);

    // PENH     
    speex_decoder_ctl(spx->dec, SPEEX_SET_ENH, &spx->penh);

    #endif

    return status;
}

pj_status_t speex_converter_close(speex_converter_data *spx)
{
    #if defined(PJMEDIA_HAS_SPEEX_CODEC) && PJMEDIA_HAS_SPEEX_CODEC!=0

    if(spx->enc)
    {
    	speex_encoder_destroy(spx->enc);
	    spx->enc = NULL;
    	speex_bits_destroy(&spx->enc_bits);
    }

    if(spx->dec)
    {
        speex_decoder_destroy(spx->dec);
        spx->dec = NULL;
        speex_bits_destroy(&spx->dec_bits);
    }

    #endif

    return PJ_SUCCESS;
}

pj_status_t speex_encode_frame(speex_converter_data *spx, const pjmedia_frame *in_frame, pjmedia_frame *out_frame)
{    
    #if defined(PJMEDIA_HAS_SPEEX_CODEC) && PJMEDIA_HAS_SPEEX_CODEC!=0

    unsigned output_buf_len = (CONF_CLOCKRATE * 20 / 1000 * CONF_CHANNELS * (CONF_BITS / 8));
    unsigned sz;
    int tx;

    if(in_frame->type != PJMEDIA_FRAME_TYPE_AUDIO)
    {
        out_frame->size = 0;
    	out_frame->buf = NULL;
	    out_frame->timestamp = in_frame->timestamp;
	    out_frame->type = in_frame->type;
        return PJ_SUCCESS;
    }

    // Flush all the bits in the struct so we can encode a new frame
    speex_bits_reset(&spx->enc_bits);

    // Encode the frame 
    tx = speex_encode_int(spx->enc, (spx_int16_t*)in_frame->buf, &spx->enc_bits);

    // Check if we need not to transmit the frame (DTX) 
    if (tx == 0) {
	    out_frame->buf = NULL;
	    out_frame->size = 0;
	    out_frame->timestamp.u64 = in_frame->timestamp.u64;
	    out_frame->type = PJMEDIA_FRAME_TYPE_NONE;
	    return PJ_SUCCESS;
    }

    // Check size. 
    sz = speex_bits_nbytes(&spx->enc_bits);
    pj_assert(sz <= output_buf_len);

    /* Copy the bits to an array of char that can be written */
    speex_bits_insert_terminator(&spx->enc_bits);
    out_frame->size = speex_bits_write(&spx->enc_bits, (char*)out_frame->buf, output_buf_len);
    out_frame->type = PJMEDIA_FRAME_TYPE_AUDIO;
    out_frame->timestamp = in_frame->timestamp;

    #endif

    return PJ_SUCCESS;
}

pj_status_t speex_decode_frame(speex_converter_data *spx, const pjmedia_frame *in_frame, pjmedia_frame *out_frame)
{
    #if defined(PJMEDIA_HAS_SPEEX_CODEC) && PJMEDIA_HAS_SPEEX_CODEC!=0

    unsigned output_buf_len = (CONF_CLOCKRATE * 20 / 1000 * CONF_CHANNELS * (CONF_BITS / 8));

    if (in_frame->type != PJMEDIA_FRAME_TYPE_AUDIO) {
	    pj_bzero(out_frame->buf, output_buf_len);
	    out_frame->size = 320;
	    out_frame->timestamp.u64 = in_frame->timestamp.u64;
	    out_frame->type = PJMEDIA_FRAME_TYPE_AUDIO;
	    return PJ_SUCCESS;
    }

    // Copy the data into the bit-stream struct 
    speex_bits_read_from(&spx->dec_bits, (char*)in_frame->buf, in_frame->size);

    // Decode the data 
    speex_decode_int(spx->dec, &spx->dec_bits, (spx_int16_t*)out_frame->buf);

    out_frame->type = PJMEDIA_FRAME_TYPE_AUDIO;
    out_frame->size = speex_bits_nbytes(&spx->dec_bits);
    pj_assert(out_frame->size <= (unsigned)output_buf_len);
    
    out_frame->size = output_buf_len;
    out_frame->timestamp.u64 = in_frame->timestamp.u64;

    #endif

    return PJ_SUCCESS;
}

#endif
